/* ssa_words.h
 * Copyright (C) 2010, Francisco Claude, all rights reserved.
 *
 * Definition of the interface to the int-based SSA that was used in
 * the word-based SSA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#ifndef SSA_WORDS_H
#define SSA_WORDS_H

#include <static_sequence.h>
#include <static_bitsequence.h>
#include <algorithm>

using namespace std;

class ssa_words
{
  public:
    ssa_words(uint * seq, uint n, bool free_text=false);
    ssa_words(FILE *fp, bool & error);
    ~ssa_words();

    bool set_static_sequence_builder(static_sequence_builder * ssb);
    bool set_static_bitsequence_builder(static_bitsequence_builder * sbb);
    bool set_samplepos(uint sample);
    bool set_samplesuff(uint sample);

    bool build_index();

    uint size();
    void print_stats();
    uint length();

    uint locate(uint * pattern, uint m, uint ** occ);
    uint count(uint * pattern, uint m);
    uint * extract(uint i, uint j);
    uint extract_pos(uint i);

    uint * rebuild();
    bool save(FILE * fp);

  protected:
    uint n,n1;
    uint sigma;
    static_sequence * bwt;
    static_bitsequence * sampled;
    uint samplepos;
    uint samplesuff;
    uint * pos_sample;
    uint * suff_sample;
    uint * sbuff;
    uint spos;

    bool built;
    bool free_text;
    uint * _seq;
    uint * _bwt;
    uint * _sa;
    static_sequence_builder * _ssb;
    static_bitsequence_builder * _sbb;

    void build_bwt();
    void build_sa();
    void sort_sa(uint ini, uint fin);
    int cmp(uint i, uint j);
    void swap(uint i, uint j);
    uint pivot(uint ini, uint fin, uint piv);
    void fill_buffer(uint i, uint j);

};
#endif
